"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDataType = exports.addParam = exports.addParams = exports.handleResult = exports.parseResult = exports.prepare = exports.stream = exports.exec = void 0;
var tslib_1 = require("tslib");
var builder_1 = require("@typed-query-builder/builder");
var sql_mssql_1 = require("@typed-query-builder/sql-mssql");
var mssql_1 = require("mssql");
require("./parsers");
function exec(access, options) {
    var _this = this;
    return function (e) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
        var request, outputFactory, output, results, e_1;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    request = new mssql_1.Request(access);
                    if (options && options.arrayMode) {
                        request.arrayRowMode = true;
                    }
                    outputFactory = sql_mssql_1.DialectMssql.output(options);
                    output = outputFactory(e);
                    addParams(request, output, options === null || options === void 0 ? void 0 : options.params);
                    _a.label = 1;
                case 1:
                    _a.trys.push([1, 3, , 4]);
                    return [4 /*yield*/, request.query(output.query)];
                case 2:
                    results = _a.sent();
                    return [2 /*return*/, parseResult(e, results, options)];
                case 3:
                    e_1 = _a.sent();
                    throw new Error(e_1 + '\n\nQuery: ' + output.query);
                case 4: return [2 /*return*/];
            }
        });
    }); };
}
exports.exec = exec;
function stream(access, options) {
    var _this = this;
    return function (e) {
        return function (listener) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var batchSize, request, outputFactory, output, accumulated_1, batchOutside_1, batchCount_1, processBatch_1;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                batchSize = (options === null || options === void 0 ? void 0 : options.streamBatchSize) || 1;
                request = new mssql_1.Request(access);
                request.stream = true;
                if (options && options.arrayMode) {
                    request.arrayRowMode = true;
                }
                outputFactory = sql_mssql_1.DialectMssql.output(options);
                output = outputFactory(e);
                addParams(request, output, options === null || options === void 0 ? void 0 : options.params);
                try {
                    request.query(output.query);
                    accumulated_1 = [];
                    batchOutside_1 = [];
                    batchCount_1 = 0;
                    processBatch_1 = function (batch) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                        var batchIndex, accumulateResult;
                        return tslib_1.__generator(this, function (_a) {
                            switch (_a.label) {
                                case 0:
                                    batchIndex = 0;
                                    _a.label = 1;
                                case 1:
                                    if (!(batchIndex < batch.length)) return [3 /*break*/, 4];
                                    return [4 /*yield*/, listener(batch[batchIndex], batchIndex, batchCount_1, batch)];
                                case 2:
                                    accumulateResult = _a.sent();
                                    if (accumulateResult !== undefined) {
                                        accumulated_1.push(accumulateResult);
                                    }
                                    _a.label = 3;
                                case 3:
                                    batchIndex++;
                                    return [3 /*break*/, 1];
                                case 4:
                                    batchCount_1++;
                                    return [2 /*return*/];
                            }
                        });
                    }); };
                    request.on('row', function (raw) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                        var result, batch;
                        return tslib_1.__generator(this, function (_a) {
                            switch (_a.label) {
                                case 0:
                                    result = parseResult(e, { recordset: [raw] }, options);
                                    batchOutside_1.push(e instanceof builder_1.QueryFirst || e instanceof builder_1.QueryFirstValue ? result : result[0]);
                                    if (!(batchOutside_1.length === batchSize)) return [3 /*break*/, 2];
                                    if (batchSize !== 1) {
                                        request.pause();
                                    }
                                    batch = batchOutside_1.slice();
                                    batchOutside_1.splice(0, batchOutside_1.length);
                                    return [4 /*yield*/, processBatch_1(batch)];
                                case 1:
                                    _a.sent();
                                    if (batchSize !== 1) {
                                        request.resume();
                                    }
                                    _a.label = 2;
                                case 2: return [2 /*return*/];
                            }
                        });
                    }); });
                    return [2 /*return*/, new Promise(function (resolve, reject) {
                            request.on('error', function (err) {
                                reject(err);
                            });
                            request.on('done', function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                return tslib_1.__generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0:
                                            if (!(batchOutside_1.length > 0)) return [3 /*break*/, 2];
                                            return [4 /*yield*/, processBatch_1(batchOutside_1.slice())];
                                        case 1:
                                            _a.sent();
                                            _a.label = 2;
                                        case 2:
                                            resolve(accumulated_1);
                                            return [2 /*return*/];
                                    }
                                });
                            }); });
                        })];
                }
                catch (e) {
                    throw new Error(e + '\n\nQuery: ' + output.query);
                }
                return [2 /*return*/];
            });
        }); };
    };
}
exports.stream = stream;
function prepare(access, options) {
    var _this = this;
    return function (e) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
        var prepared, outputFactory, output, defaults, paramName, paramIndex, paramTypeGiven, paramValue, paramType, preparedQuery;
        return tslib_1.__generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    prepared = new mssql_1.PreparedStatement(access);
                    if (options && options.arrayMode) {
                        prepared.arrayRowMode = true;
                    }
                    outputFactory = sql_mssql_1.DialectMssql.output(options);
                    output = outputFactory(e);
                    defaults = {};
                    for (paramName in output.paramIndices) {
                        paramIndex = output.paramIndices[paramName];
                        paramTypeGiven = output.paramTypes[paramIndex];
                        paramValue = output.params[paramIndex];
                        paramType = getDataType(paramValue, paramTypeGiven);
                        defaults[paramName] = paramValue;
                        prepared.input(paramName, paramType);
                    }
                    return [4 /*yield*/, prepared.prepare(output.query)];
                case 1:
                    _a.sent();
                    preparedQuery = {
                        exec: function (params) {
                            return tslib_1.__awaiter(this, void 0, void 0, function () {
                                var result;
                                return tslib_1.__generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, prepared.execute(tslib_1.__assign(tslib_1.__assign(tslib_1.__assign({}, defaults), ((options === null || options === void 0 ? void 0 : options.params) || {})), params))];
                                        case 1:
                                            result = _a.sent();
                                            return [2 /*return*/, parseResult(e, result, options)];
                                    }
                                });
                            });
                        },
                        release: function () {
                            return tslib_1.__awaiter(this, void 0, void 0, function () {
                                return tslib_1.__generator(this, function (_a) {
                                    switch (_a.label) {
                                        case 0: return [4 /*yield*/, prepared.unprepare()];
                                        case 1:
                                            _a.sent();
                                            return [2 /*return*/];
                                    }
                                });
                            });
                        },
                    };
                    return [2 /*return*/, preparedQuery];
            }
        });
    }); };
}
exports.prepare = prepare;
function parseResult(expr, iresult, options) {
    var e_2, _a;
    var result = sql_mssql_1.DialectMssql.getResult(expr, handleResult(expr, iresult, options));
    if (options && (options.detectJson || options.detectDate || options.detectSmallDateTime || options.detectDateTime || options.detectDateTime2 || options.detectDateTimeOffset || options.detectAllDates)) {
        var traverse_1 = function (value, onValue) {
            value = onValue(value);
            if (builder_1.isArray(value)) {
                for (var i = 0; i < value.length; i++) {
                    value[i] = traverse_1(value[i], onValue);
                }
            }
            else if (builder_1.isPlainObject(value)) {
                for (var prop in value) {
                    value[prop] = traverse_1(value[prop], onValue);
                }
            }
            return value;
        };
        if (options.detectJson) {
            result = traverse_1(result, function (value) {
                if (builder_1.isString(value) && value.match(/^({|true$|false$|\[|null$|[+-]?\d|")/)) {
                    try {
                        return JSON.parse(value);
                    }
                    catch (e) { }
                }
                return value;
            });
        }
        var dateDetectors_2 = [
            {
                detect: options.detectDate || options.detectAllDates,
                length: 10,
                regex: /^\d{4}-\d\d-\d\d$/,
                toDate: function (value) { return new Date(value + 'T00:00:00'); },
            },
            {
                detect: options.detectSmallDateTime || options.detectAllDates,
                length: 19,
                regex: /^\d{4}-\d\d-\d\d(T|\s)\d\d:\d\d:\d\d?$/,
                toDate: function (value) { return new Date(value); },
            },
            {
                detect: options.detectDateTime || options.detectAllDates,
                length: 23,
                regex: /^\d{4}-\d\d-\d\d(T|\s)\d\d:\d\d:\d\d\.\d{3}?$/,
                toDate: function (value) { return new Date(value); },
            },
            {
                detect: options.detectDateTime2 || options.detectAllDates,
                length: 27,
                regex: /^\d{4}-\d\d-\d\d(T|\s)\d\d:\d\d:\d\d\.\d{7}?$/,
                toDate: function (value) { return new Date(value); },
            },
            {
                detect: options.detectDateTimeOffset || options.detectAllDates,
                length: 34,
                regex: /^\d{4}-\d\d-\d\d(T|\s)\d\d:\d\d:\d\d\.\d{7}\s[+-]\d\d:\d\d?$/,
                toDate: function (value) { return new Date(value.replace(/\s([-+])/, '$1')); },
            }
        ];
        if (dateDetectors_2.some(function (d) { return d.detect; })) {
            try {
                for (var dateDetectors_1 = tslib_1.__values(dateDetectors_2), dateDetectors_1_1 = dateDetectors_1.next(); !dateDetectors_1_1.done; dateDetectors_1_1 = dateDetectors_1.next()) {
                    var detector = dateDetectors_1_1.value;
                    if (builder_1.isFunction(detector.detect)) {
                        detector.toDate = detector.detect;
                    }
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (dateDetectors_1_1 && !dateDetectors_1_1.done && (_a = dateDetectors_1.return)) _a.call(dateDetectors_1);
                }
                finally { if (e_2) throw e_2.error; }
            }
            result = traverse_1(result, function (value) {
                var e_3, _a;
                if (builder_1.isString(value)) {
                    try {
                        for (var dateDetectors_3 = tslib_1.__values(dateDetectors_2), dateDetectors_3_1 = dateDetectors_3.next(); !dateDetectors_3_1.done; dateDetectors_3_1 = dateDetectors_3.next()) {
                            var detector = dateDetectors_3_1.value;
                            if (value.length === detector.length && value.match(detector.regex)) {
                                return detector.toDate(value);
                            }
                        }
                    }
                    catch (e_3_1) { e_3 = { error: e_3_1 }; }
                    finally {
                        try {
                            if (dateDetectors_3_1 && !dateDetectors_3_1.done && (_a = dateDetectors_3.return)) _a.call(dateDetectors_3);
                        }
                        finally { if (e_3) throw e_3.error; }
                    }
                }
                return value;
            });
        }
    }
    return (options === null || options === void 0 ? void 0 : options.affectedCount) ? { affected: iresult.rowsAffected[0], result: result }
        : result;
}
exports.parseResult = parseResult;
function handleResult(expr, result, options) {
    if (expr instanceof builder_1.QueryFirst) {
        return result.recordset[0] || null;
    }
    if (expr instanceof builder_1.QueryFirstValue) {
        for (var prop in result.recordset[0]) {
            return result.recordset[0][prop];
        }
    }
    if (expr instanceof builder_1.QueryList) {
        return result.recordset.map(function (r) { return r.item; });
    }
    if (expr instanceof builder_1.QueryJson) {
        return result.recordset[0];
    }
    return result.recordset;
}
exports.handleResult = handleResult;
function addParams(request, out, params) {
    for (var paramName in out.paramIndices) {
        if (!(paramName in params)) {
            var paramIndex = out.paramIndices[paramName];
            addParam(request, out, paramName, out.params[paramIndex]);
        }
    }
    if (params) {
        for (var paramName in params) {
            addParam(request, out, paramName, params[paramName]);
        }
    }
}
exports.addParams = addParams;
function addParam(request, out, paramName, paramValue) {
    var paramIndex = out.paramIndices[paramName];
    var paramTypeGiven = out.paramTypes[paramIndex];
    var paramType = getDataType(paramValue, paramTypeGiven);
    request.input(paramName, paramType, paramValue);
}
exports.addParam = addParam;
function getDataType(value, givenType) {
    var dataType = givenType || builder_1.getDataTypeFromValue(value);
    var dataTypeName = builder_1.getDataTypeFromInput(dataType);
    var meta = builder_1.getDataTypeMeta(dataType);
    switch (dataTypeName) {
        case 'BOOLEAN':
        case 'BIT':
            return mssql_1.TYPES.Bit;
        case 'TINYINT':
            return mssql_1.TYPES.TinyInt;
        case 'SMALLINT':
            return mssql_1.TYPES.SmallInt;
        case 'MEDIUMINT':
        case 'INT':
            return mssql_1.TYPES.Int;
        case 'BIGINT':
            return mssql_1.TYPES.BigInt;
        case 'FLOAT':
        case 'DOUBLE':
            return mssql_1.TYPES.Float();
        case 'DECIMAL':
            return mssql_1.TYPES.Decimal(meta.totalDigits, meta.fractionDigits);
        case 'NUMERIC':
            return mssql_1.TYPES.Numeric(meta.totalDigits, meta.fractionDigits);
        case 'MONEY':
            return mssql_1.TYPES.Money;
        case 'SMALLMONEY':
            return mssql_1.TYPES.SmallMoney;
        case 'NCHAR':
            return mssql_1.TYPES.NChar(meta.length);
        case 'CHAR':
            return mssql_1.TYPES.Char(meta.length);
        case 'NVARCHAR':
            return mssql_1.TYPES.NVarChar(meta.length);
        case 'VARCHAR':
            return mssql_1.TYPES.VarChar(meta.length);
        case 'NTEXT':
            return mssql_1.TYPES.NText();
        case 'TEXT':
            return mssql_1.TYPES.Text();
        case 'TIMESTAMP':
            return mssql_1.TYPES.DateTime2();
        case 'DATE':
            return mssql_1.TYPES.Date();
        case 'TIME':
            return mssql_1.TYPES.Time();
        case 'UUID':
            return mssql_1.TYPES.UniqueIdentifier();
        case 'CIDR':
            return mssql_1.TYPES.VarChar(45);
        case 'INET':
            return mssql_1.TYPES.VarChar(45);
        case 'MACADDR':
            return mssql_1.TYPES.VarChar(17);
        case 'BINARY':
            return mssql_1.TYPES.Binary;
        case 'VARBINARY':
            return mssql_1.TYPES.VarBinary(meta.length);
        case 'BLOB':
            return mssql_1.TYPES.VarBinary(mssql_1.MAX);
        case 'JSON':
            return mssql_1.TYPES.NVarChar(mssql_1.MAX);
        case 'XML':
            return mssql_1.TYPES.Xml;
        case 'BOX':
        case 'CIRCLE':
        case 'LINE':
        case 'SEGMENT':
        case 'PATH':
        case 'POLYGON':
        case 'POINT':
        case 'GEOMETRY':
            return meta.srid === 0 ? mssql_1.TYPES.Geometry : mssql_1.TYPES.Geography;
    }
    return mssql_1.TYPES.NVarChar(mssql_1.MAX);
}
exports.getDataType = getDataType;
//# sourceMappingURL=index.js.map